from colormath.color_objects import sRGBColor, HSVColor
from colormath.color_conversions import convert_color
import numpy as np


class ColorSelectorBase:
    @staticmethod
    def _make_cols(n, hue):
        return [
            convert_color(
                HSVColor(hue, i/(n-1), 1),
                sRGBColor
            ).get_rgb_hex()
            for i in range(n)
        ]

    def __call__(self, v):
        pos = 0
        while pos < self._n_tick and self._tick[pos] <= v:
            pos += 1
        return self._cols[pos]


class color_selector(ColorSelectorBase):
    def __init__(self, values, n, hue=0.9, min_value=-np.Inf,
                 max_value=np.Inf, desc=False):
        self._n_cols = n
        self._n_tick = n - 1
        self._cols = ColorSelectorBase._make_cols(n, hue)
        if desc:
            self._cols = list(reversed(self._cols))
        m = max(min_value, min(values))
        M = min(max_value, max(values))
        self._tick = np.linspace(m, M, self.n_cols+1)[1:-1]


class color_selector_tick(ColorSelectorBase):
    def __init__(self, ticks, hue=0.9, reverse=False):
        self._n_cols = len(ticks) + 1
        self._n_tick = len(ticks)
        self._cols = ColorSelectorBase._make_cols(self._n_cols, hue)
        if reverse:
            self._cols = list(reversed(self._cols))
        self._tick = ticks


class color_selector_p(ColorSelectorBase):
    def __init__(self, values, n, hue=0.9):
        self._n_tick = n - 1
        self._n_cols = n
        self._cols = ColorSelectorBase._make_cols(self._n_cols, hue)
        ps = np.linspace(0, 100, self._n_cols+1)[1:-1]
        self._tick = np.percentile(values, ps)



import time
import sys

_time = None


def start():
    global _time
    _time = time.time()


def end(dest=sys.stderr):
    val = time.time() - _time
    if dest is None:
        return val
    else:
        dest.write(f"elapsed: {val}")



import itertools as it
import matplotlib.pyplot as plt
from h3 import h3
import numpy as np
import functools as ft
import tilemapbase as tmb
import folium
from collections import namedtuple
import geojson as gj


def _flatten(lss):
    return ft.reduce(lambda x, y: x+y, lss)


def _swap(ls):
    return [
        (p[1], p[0])
        for p in ls
    ]


def _extent(ids):
    RetExtent = namedtuple(
        "RetExtent", "extent min_lng max_lng min_lat max_lat")

    ids_u = ids.unique()
    latlngs = _flatten(map(h3.h3_to_geo_boundary, ids_u))
    lats = [x[0] for x in latlngs]
    lngs = [x[1] for x in latlngs]
    extent = tmb.Extent.from_lonlat(
        min(lngs), max(lngs),
        min(lats), max(lats)
    )
    extent.to_aspect(1)
    return RetExtent(extent, min(lngs), max(lngs), min(lats), max(lats))


def extent_minmax(lat, lng):
    ret = tmb.Extent.from_lonlat(
        lng[0], lng[1],
        lat[0], lat[1]
    )
    ret.to_aspect(1)
    return ret


LatLng = namedtuple("LatLng", "lat lng")


def extent_corner(ll, ur):
    return extent_minmax(
        (ll.lat, ur.lat),
        (ll.lng, ur.lng)
    )


def extent(ids):
    return _extent(ids).extent


def n_extent(ids, n):
    extent_ = _extent(ids)
    lng_width = (extent_.max_lng - extent_.min_lng) / n
    lat_width = (extent_.max_lat - extent_.min_lat) / n
    lng_ticks = [
        extent_.min_lng + i * lng_width
        for i in range(n+1)
    ]
    lat_ticks = [
        extent_.min_lat + i * lat_width
        for i in range(n+1)
    ]
    MinMax = namedtuple("MinMax", "min max")
    lng_bounds = map(lambda x: MinMax(*x), zip(lng_ticks[:-1], lng_ticks[1:]))
    lat_bounds = map(lambda x: MinMax(*x), zip(lat_ticks[:-1], lat_ticks[1:]))

    Bound = namedtuple("Bound", "longitude latitude")
    bounds = list(map(lambda x: Bound(*x), it.product(lng_bounds, lat_bounds)))
    extents = [
        tmb.Extent.from_lonlat(
            bound.longitude.min, bound.longitude.max,
            bound.latitude.min, bound.latitude.max
        ) for bound in bounds
    ]

    def search_position(id):
        lat, lng = h3.h3_to_geo(id)
        xpos = int((lng - extent_.min_lng) // lng_width)
        ypos = int((lat - extent_.min_lat) // lat_width)
        # import pdb
        # pdb.set_trace()
        return n*xpos + ypos
    search_position_v = np.vectorize(search_position)
    positions = search_position_v(ids)

    Return = namedtuple("Return", "extents positions")
    return Return(extents, positions)


def draw(df, id_col, val_col, extent, color_selector, figsize=(8, 8), dpi=100, width=600, alpha=0.8, axis_visible=False):
    fig, ax = plt.subplots(figsize=figsize, dpi=dpi)
    ax.xaxis.set_visible(axis_visible)
    ax.yaxis.set_visible(axis_visible)

    t = tmb.tiles.build_OSM()
    plotter = tmb.Plotter(extent, t, width=width)
    plotter.plot(ax, t)

    n = len(df)
    for i in range(n):
        id = df[id_col].iloc[i]
        val = df[val_col].iloc[i]
        color = color_selector(val)
        vts = _swap(h3.h3_to_geo_boundary(id))
        xys = [tmb.project(*x) for x in vts]
        poly = plt.Polygon(xys, fc=color, alpha=alpha)
        ax.add_patch(poly)

    return fig, ax


def draw_folium(df, id_col, val_col, color_selector, zoom_start, popups=["val", "link"], draw_line=False, label_col=None, latlng_popup=True):
    n = len(df)
    lats, lngs = 0, 0
    polys = {}
    for i in range(n):
        id = df[id_col].iloc[i]
        val = df[val_col].iloc[i]
        color = color_selector(val)
        vts = h3.h3_to_geo_boundary(id)
        for v in vts:
            lats += v[0]
            lngs += v[1]
        polys[i] = folium.Polygon(
            locations=vts,
            color="#080016",
            fill=True,
            fill_color=color,
            fill_opacity=0.8,
            weight=0.3 if draw_line else 0
        )
        pop = []
        if "val" in popups:
            pop.append(val)
        if "link" in popups:
            pos = h3.h3_to_geo(id)
            pop.append(
                f'<a href="https://www.google.com/maps/search/{pos[0]},+{pos[1]}">link</a>')
        if "h3" in popups:
            pop.append(id)
        if "latlng" in popups:
            pos = h3.h3_to_geo(id)
            pop.append(pos)
        if label_col is not None:
            pop.append(df[label_col].iloc[i])
        if len(pop) > 0:
            polys[i].add_child(folium.Popup("\n".join(map(str, pop))))
    lat, lng = lats/(6*n), lngs/(6*n)
    fmap = folium.Map(
        location=[lat, lng],
        zoom_start=zoom_start
    )
    for k, v in polys.items():
        v.add_to(fmap)
    if latlng_popup:
        folium.LatLngPopup().add_to(fmap)
    return fmap


def extract_part(df, id_col, min_latlng, max_latlng):
    min_lat, min_lng = min_latlng
    max_lat, max_lng = max_latlng

    def selector(h3id):
        lat, lng = h3.h3_to_geo(h3id)
        return min_lat <= lat <= max_lat and min_lng <= lng <= max_lng
    v = np.vectorize(selector)
    return df.loc[v(df[id_col])]


def drawp(df, poly_col, val_col, extent, color_selector,
          figsize=(8, 8), dpi=100, width=600, alpha=0.8):
    fig, ax = plt.subplots(figsize=figsize, dpi=dpi)
    ax.xaxis.set_visible(False)
    ax.yaxis.set_visible(False)

    t = tmb.tiles.build_OSM()
    plotter = tmb.Plotter(extent, t, width=width)
    plotter.plot(ax, t)

    n = len(df)
    for i in range(n):
        val = df[val_col].iloc[i]
        color = color_selector(val)
        vts = df[poly_col].iloc[i]["coordinates"][0]
        xys = [tmb.project(*x) for x in vts]
        poly = plt.Polygon(xys, fc=color, alpha=alpha)
        ax.add_patch(poly)

    return fig, ax


def extentp(polys):
    vts = _flatten([gj.loads(p)["coordinates"][0] for p in polys])
    lngs = [x[0] for x in vts]
    lats = [x[1] for x in vts]
    extent = tmb.Extent.from_lonlat(
        min(lngs), max(lngs),
        min(lats), max(lats)
    )
    extent.to_aspect(1)
    return extent



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd


def percent_in_rows(df):
    sums = df.apply(sum, axis=1)
    df2 = df.copy()
    for i in range(len(df)):
        df2.iloc[i, :] = df.iloc[i, :] / sums.iloc[i] * 100
    return df2


def cumplot_help(df, x, y, category, percent):
    df1 = df.loc[:, [x, y, category]]
    df2 = df1.pivot(index=x, columns=category, values=y).fillna(0)

    if percent:
        df2 = percent_in_rows(df2)
    cum1 = np.cumsum(df2, axis=1)
    zeros = pd.Series(np.zeros(len(cum1)), index=cum1.index)
    cum2 = pd.concat([zeros, cum1], axis=1).iloc[:, :-1]
    cum2.columns = df2.columns
    return df2, cum2


def do_plot(x, y, offset, horizontal=False, ax=None):
    obj = plt if ax is None else ax
    if horizontal:
        return obj.barh(x, y, left=offset)
    else:
        return obj.bar(x, y, bottom=offset)


def cumplot(df, x, y, category, horizontal=False, percent=False, ax=None, **kwargs):
    df2, cum2 = cumplot_help(df, x, y, category, percent)
    ps = [None] * len(df2.columns)
    for i in range(len(df2.columns)):
        ps[i] = do_plot(df2.index, df2.iloc[:, i],
                        cum2.iloc[:, i], horizontal, ax)
    obj = plt if ax is None else ax
    obj.legend(map(lambda p: p[0], ps), df2.columns, **kwargs)


def abline(a, b, axes):
    x_vals = np.array(axes.get_xlim())
    y_vals = a * x_vals + b
    axes.plot(x_vals, y_vals, '-')



# mapのコピーライト
# open street map
copyright_osm = '&copy; <a href="http://osm.org/copyright">OpenStreetMap</a> contributors'

# Stamen Toner
copyright_st = '&copy; ' \
            'Map tiles by <a href="http://stamen.com">Stamen Design</a>,' \
            ' under <a href="http://creativecommons.org/licenses/by/3.0">CC BY 3.0</a>.' \
            'Data by <a href="http://openstreetmap.org">OpenStreetMap</a>,' \
            'under <a href="http://www.openstreetmap.org/copyright">ODbL</a>.'



import subprocess as sp
import sys


def make_gif(files, output, interval=10):
    cmd = [
        "convert",
        "-layers",
        "optimize",
        "-loop",
        "0",
        "-delay",
        str(interval)
    ] + list(files) + [
        output
    ]
    r = sp.run(cmd, stdout=sp.PIPE, stderr=sp.PIPE)
    o = r.stdout.decode("utf-8")
    e = r.stderr.decode("utf-8")
    print(o)
    if len(e) > 0:
        msg = f"{e}\n{' '.join(cmd)}\n"
        sys.stderr.write(msg)



import tilemapbase as tmb
import matplotlib.pyplot as plt
import matplotlib.lines as lines
import folium
from .maputil import copyright_osm
import numpy as np
from geopy.distance import distance
import re

tmb.init(create=True)


def _extend(m, M, p):
    w = M - m
    return m - p*w, M + p*w


def _expand(ex, p):
    xmin, xmax = _extend(ex.xmin, ex.xmax, p)
    ymin, ymax = _extend(ex.ymin, ex.ymax, p)
    return tmb.Extent(xmin, xmax, ymin, ymax)


def _widen(ex, p):
    return tmb.Extent(
        *_extend(ex.xmin, ex.xmax, p),
        ex.ymin, ex.ymax
    )


def _heighten(ex, p):
    return tmb.Extent(
        ex.xmin, ex.xmax,
        *_extend(ex.ymin, ex.ymax, p)
    )


def _adjust(ex, w, h):
    # Extent.xrange, yrange returns min and max in skewed way
    xmin, xmax = ex.xrange
    ymax, ymin = ex.yrange
    m = ymax - ymin
    n = xmax - xmin
    if h < w:
        p1 = w/h
        p2 = m/n
        return _widen(ex, p1 * p2)
    elif w < h:
        p1 = h/w
        p2 = n/m
        return _heighten(ex, p1 * p2)
    elif m < n:
        return _heighten(ex, n/m)
    elif n < m:
        return _widen(ex, m/n)
    else:
        return ex

    return tmb.Extent(xmin, xmax, ymin, ymax)


def _get_column(df, candidates):
    clns = [x.lower() for x in df.columns]
    candidates = [c.lower() for c in candidates if c is not None]
    pos = -1
    for c in candidates:
        if c in clns:
            pos = clns.index(c)
            break
    if pos == -1:
        raise RuntimeError(f"{candidates} is not found in {clns}")
    return df.iloc[:, pos]


def _iterable(x):
    try:
        iter(x)
    except TypeError:
        return False
    return True


def draw(df,
         latitude=None, longitude=None,
         p_size=1, p_color="#000000", p_popup=None,
         l_size=1, l_color="#000000", l_popup=None,
         output_format="png",
         **kwargs):
    lats = _get_column(df, [latitude, "latitude", "lat"]).values
    lngs = _get_column(df, [longitude, "longitude", "lon", "lng"]).values

    if type(p_size) is str:
        ps = _get_column(df, [p_size])
    elif type(p_size) is float or type(p_size) is int:
        ps = [p_size] * len(df)
    elif type(p_size) is list:
        ps = p_size
    elif _iterable(p_size):
        ps = list(p_size)
    else:
        raise ValueError(f"{p_size} is given for p_size")

    if type(p_color) is str and re.match(r"^#[0-9a-fA-F]{6}$", p_color):
        pc = [p_color] * len(df)
    elif type(p_color) is str:
        pc = _get_column(df, [p_color])
    elif type(p_color) is list:
        pc = p_color
    elif _iterable(p_color):
        pc = list(p_color)
    else:
        raise ValueError(f"{p_color} is given for p_color")

    if type(l_size) is str:
        ls = _get_column(df, [l_size])
    elif type(l_size) is float or type(l_size) is int:
        ls = [l_size] * (len(df) - 1)
    elif type(l_size) is list:
        ls = l_size
    elif _iterable(l_size):
        ls = list(l_size)
    else:
        raise ValueError(f"{l_size} is given for l_size")

    if type(l_color) is str and re.match(r"^#[0-9a-fA-F]{6}$", l_color):
        lc = [l_color] * (len(df) - 1)
    elif type(l_color) is str:
        lc = _get_column(df, [l_color])
    elif type(l_color) is list:
        lc = l_color
    elif _iterable(l_color):
        lc = list(l_color)
    else:
        raise ValueError(f"{l_color} is given for l_color")

    if output_format == "png":
        return _draw_png(
            df,
            lats, lngs,
            ps, pc,
            ls, lc,
            **kwargs
        )
    elif output_format == "html":
        return _draw_html(
            df,
            lats, lngs,
            ps, pc, p_popup,
            ls, lc, l_popup,
            ** kwargs
        )
    else:
        raise ValueError(
            f"'{output_format}' is not valid for output_format. It should be 'png' or 'html'")


def _draw_png(df,
              lats, lngs,
              p_size, p_color,
              l_size, l_color,
              figsize=(8, 8), dpi=100,
              axis_visible=False,
              padding=0.03,
              adjust=True):
    ex1 = tmb.Extent.from_lonlat(
        min(lngs), max(lngs),
        min(lats), max(lats)
    )
    ex2 = _expand(ex1, padding)
    extent = ex2.to_aspect(figsize[0]/figsize[1],
                           shrink=False) if adjust else ex2

    fig, ax = plt.subplots(figsize=figsize, dpi=dpi)
    ax.xaxis.set_visible(axis_visible)
    ax.yaxis.set_visible(axis_visible)

    t = tmb.tiles.build_OSM()
    plotter = tmb.Plotter(extent, t,
                          width=figsize[0] * 100,
                          height=figsize[1] * 100)
    plotter.plot(ax, t)

    ps = [tmb.project(x, y) for x, y in zip(lngs, lats)]
    xs = [p[0] for p in ps]
    ys = [p[1] for p in ps]
    n = len(df)
    for i in range(n-1):
        l2 = lines.Line2D(xs[i:(i+2)], ys[i:(i+2)],
                          linewidth=l_size[i], color=l_color[i])
        ax.add_line(l2)
    for i in range(n):
        x, y = ps[i]
        ax.plot(x, y, marker=".", markersize=p_size[i], color=p_color[i])

    return fig, ax


def _draw_html(df,
               lats, lngs,
               p_size, p_color, p_popup,
               l_size, l_color, l_popup,
               zoom_start=15, width=800, height=800):
    n = len(df)
    mlat = np.mean(lats)
    mlng = np.mean(lngs)
    fmap = folium.Map(
        location=[mlat, mlng],
        attr=copyright_osm,
        width=width, height=height,
        zoom_start=zoom_start
    )

    for i in range(n):
        x, y = lngs[i], lats[i]
        folium.Circle(
            (y, x),
            color=p_color[i],
            fill=True,
            popup=p_popup[i] if p_popup is not None else None,
            radius=p_size[i],
            weight=0
        ).add_to(fmap)
    for i in range(n-1):
        x, y = lngs[i], lats[i]
        nx, ny = lngs[i+1], lats[i+1]
        col = l_color[i]
        folium.PolyLine(
            locations=[(y, x), (ny, nx)],
            color=col,
            weight=l_size[i],
            popup=l_popup[i] if l_popup is not None else None
        ).add_to(fmap)

    return fmap



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    "colormath",
    "folium",
    "geojson",
    "geopy",
    "h3",
    "matplotlib",
    "numpy",
    "pandas",
    "pyproj",
    "tilemapbase"
]

setup_requirements = []

test_requirements = []

setup(
    author="Yasunori Horikoshi",
    author_email='yasunori_horikoshi@datawise.co.jp',
    python_requires='>=3.5',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="tools for data science",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='dtws1',
    name='dtws1',
    packages=find_packages(include=['dtws1', 'dtws1.*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/hotoku-dtws/dtws1',
    version='0.2.4',
    zip_safe=False,
)
